﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Threading;
using Infragistics.Controls.Grids;
using Infragistics.Olap;
using Infragistics.Olap.Data;
using ExecutionContext = Infragistics.Olap.ExecutionContext;

namespace SaveLoadViewsSample
{
    public class SavedPivotGridView
    {
        #region SavedInformation

        public string Timestamp { get; set; }

        public int SelectedCubeIndex { get; set; }

        //unique names of each row, column, filter and measure that will be saved
        public List<string> RowHierarchies { get; set; }
        public List<string> ColumnHierarchies { get; set; }
        public List<string> FilterHierarchies { get; set; }
        public List<string> Measures { get; set; }

        //unique name of each member to save and its IsExpanded property
        public List<CustomDictItem> FilterMembersToExpand { get; set; }
        //unique name of each member to save and its IsSelected property
        public List<CustomDictItem> FilterMembersToLoad { get; set; }
        public int MeasureListIndex { get; set; }
        public MeasureListLocation MeasureListLocation { get; set; }

        [DefaultValue(0)]
        public int HierarchiesLoaded { get; set; }
        #endregion SavedInformation

        #region SaveMethods

        public void SaveView(XamPivotGrid pivotGrid)
        {
            //check if pivotGrid actually has information that can be saved
            if (pivotGrid.DataSource == null)
                return;

            if (pivotGrid.DataSource.Cube == null)
                return;

            //give a name for the saved view - default is the time stamp of the save moment
            this.Timestamp = DateTime.Now.ToString();

            //save the selected cube
            this.SelectedCubeIndex = pivotGrid.DataSource.Cubes.IndexOf(pivotGrid.DataSource.Cube);

            this.MeasureListIndex = pivotGrid.DataSource.MeasureListViewModel.Index;
            this.MeasureListLocation = pivotGrid.DataSource.MeasureListViewModel.MeasureListLocation;

            //save the row hierarchies
            this.RowHierarchies = new List<string>();
            foreach (IAreaItemViewModel item in pivotGrid.DataSource.Rows)
            {
                IFilterViewModel FilterViewModel = item as IFilterViewModel;
                if (FilterViewModel != null)
                {
                    this.RowHierarchies.Add(FilterViewModel.Hierarchy.UniqueName);
                }
            }

            //save the column hierarchies
            this.ColumnHierarchies = new List<string>();
            foreach (IAreaItemViewModel item in pivotGrid.DataSource.Columns)
            {
                IFilterViewModel FilterViewModel = item as IFilterViewModel;
                if (FilterViewModel != null)
                {
                    this.ColumnHierarchies.Add(FilterViewModel.Hierarchy.UniqueName);
                }
            }

            //save the filter hierarchies
            this.FilterHierarchies = new List<string>();
            foreach (IAreaItemViewModel item in pivotGrid.DataSource.Filters)
            {
                IFilterViewModel FilterViewModel = item as IFilterViewModel;
                if (FilterViewModel != null)
                {
                    this.FilterHierarchies.Add(FilterViewModel.Hierarchy.UniqueName);
                }
            }

            //save the measures
            this.Measures = new List<string>();
            foreach (IAreaItemViewModel item in pivotGrid.DataSource.Measures)
            {
                IMeasureViewModel measureViewModel = item as IMeasureViewModel;
                if (measureViewModel != null)
                {
                    this.Measures.Add(measureViewModel.Measure.UniqueName);
                }
            }

            //save the filter members expand states. 
            //this is needed so we can expand the levels of the tree that contains the filter members.
            //only if we have this info we can go through all filter members in the filter hierarchies
            //and restore their selection states
            FilterMembersToExpand = new List<CustomDictItem>();
            foreach (IFilterViewModel filterViewModel in pivotGrid.DataSource.Filters)
            {
                if (filterViewModel.FilterMembers != null)
                    foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                        SaveExpansionStateOfFilterMembers(filterMember);
            }
            foreach (PivotHeaderCell cell in pivotGrid.GridLayout.RowHeaderCells)
            {
                CustomDictItem item = new CustomDictItem(cell.Member.UniqueName, cell.IsExpanded);

                if (!CustomDictContains(this.FilterMembersToExpand, item.UniqueName))
                {
                    this.FilterMembersToExpand.Add(item);
                }
                else
                {
                    GetCustomDictItem(this.FilterMembersToExpand, item.UniqueName).Value = true;
                }
            }
            foreach (PivotHeaderCell cell in pivotGrid.GridLayout.ColumnHeaderCells)
            {
                CustomDictItem item = new CustomDictItem(cell.Member.UniqueName, cell.IsExpanded);

                if (!CustomDictContains(this.FilterMembersToExpand, item.UniqueName))
                {
                    this.FilterMembersToExpand.Add(item);
                }
                else
                {
                    GetCustomDictItem(this.FilterMembersToExpand, item.UniqueName).Value = true;
                }
            }

            //save the filter members selected states for column, row and filters hierarchies.
            //this is needed so we can restore the selected filter members 
            //from the filter menu of each hierarchy added to the grid
            FilterMembersToLoad = new List<CustomDictItem>();
            for (int i = 0; i < pivotGrid.DataSource.Columns.Count; i++)
            {
                IFilterViewModel filterViewModel = pivotGrid.DataSource.Columns[i] as IFilterViewModel;

                if (filterViewModel != null)
                {
                    if (filterViewModel.FilterMembers != null)
                        foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                            SaveSelectionStateOfFilterMembers(filterMember);
                }
            }
            for (int i = 0; i < pivotGrid.DataSource.Rows.Count; i++)
            {
                IFilterViewModel filterViewModel = pivotGrid.DataSource.Rows[i] as IFilterViewModel;

                if (filterViewModel != null)
                {
                    if (filterViewModel.FilterMembers != null)
                        foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                            SaveSelectionStateOfFilterMembers(filterMember);
                }
            }
            foreach (IFilterViewModel filterViewModel in pivotGrid.DataSource.Filters)
            {
                if (filterViewModel.FilterMembers != null)
                    foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                        SaveSelectionStateOfFilterMembers(filterMember);
            }
        }

        //recusive method to save the expansion states of the filter members
        private void SaveExpansionStateOfFilterMembers(IFilterMember filterMember)
        {
            if (filterMember.FilterMembers != null)
            {
                foreach (IFilterMember subFilterMember in filterMember.FilterMembers)
                {
                    SaveExpansionStateOfFilterMembers(subFilterMember);
                }
            }

            CustomDictItem item = new CustomDictItem(filterMember.Member.UniqueName, filterMember.IsExpanded);

            if (!CustomDictContains(this.FilterMembersToExpand, item.UniqueName))
            {
                this.FilterMembersToExpand.Add(item);
            }
            else
            {
                GetCustomDictItem(this.FilterMembersToExpand, item.UniqueName).Value = true;
            }
        }

        //recusive method to save the selection states of the filter members
        private void SaveSelectionStateOfFilterMembers(IFilterMember filterMember)
        {
            if (filterMember.FilterMembers != null)
            {
                foreach (IFilterMember subFilterMember in filterMember.FilterMembers)
                {
                    SaveSelectionStateOfFilterMembers(subFilterMember);
                }
            }

            if (filterMember.IsSelected.HasValue)
            {
                CustomDictItem item =
                    new CustomDictItem(filterMember.Member.UniqueName, filterMember.IsSelected.Value);

                if (!CustomDictContains(this.FilterMembersToLoad, item.UniqueName))
                {
                    this.FilterMembersToLoad.Add(item);
                }
                else
                {
                    GetCustomDictItem(this.FilterMembersToLoad, item.UniqueName).Value = true;
                }
            }
        }

        #endregion SaveMethods

        #region LoadMethods

        public void Fill(XamPivotGrid pivotGrid)
        {
            //this flag facilitates the timing of the load events
            StateLoadComplete = false;

            //select the stored cube index
            pivotGrid.DataSource.Cube = pivotGrid.DataSource.Cubes[this.SelectedCubeIndex];

            //enable deferred layout update so all restore procedures are executed at once. 
            //otherwise the UI will be updated after each command and thus it will be slow
            pivotGrid.DataSource.DeferredLayoutUpdate = true;

            //clear and then load the saved row, column and filter hierarchies and then the saved measures
            pivotGrid.DataSource.Rows.Clear();
            pivotGrid.DataSource.Columns.Clear();
            pivotGrid.DataSource.Measures.Clear();
            pivotGrid.DataSource.Filters.Clear();

            while (HierarchiesLoaded < RowHierarchies.Count)
            {
                Fill_ProcessDimensions(pivotGrid.DataSource.Cube.Dimensions, pivotGrid);
            }

            // restore the location and the index at which measure list item appears
            pivotGrid.DataSource.MeasureListLocation = this.MeasureListLocation;
            pivotGrid.DataSource.MeasureListIndex = this.MeasureListIndex;

            foreach (IMeasure measure in pivotGrid.DataSource.Cube.Measures)
            {
                if (this.Measures.Contains(measure.UniqueName))
                {
                    IMeasureViewModel mvm = pivotGrid.DataSource.CreateMeasureViewModel(measure);
                    pivotGrid.DataSource.Measures.Add(mvm);
                }
            }

            //attach to the layoutloaded event so expansion of all hierarchy levels 
            //can be executed after the UI is updated
            if (!eventAttached)
                pivotGrid.LayoutLoaded += pivotGrid_LayoutLoaded;

            //turn of the deferred layout update and force an UI udpate
            pivotGrid.DataSource.DeferredLayoutUpdate = false;
            //reset hierarchies loaded to zero
            HierarchiesLoaded = 0;
        }

        void Fill_ProcessDimensions(OlapElementCollection<IDimension> pivotDimensions, XamPivotGrid pivotGrid)
        {
            int LoadCounter = 0;

            foreach (IDimension dim in pivotDimensions)
            {
                foreach (IHierarchy hierarchy in dim.Hierarchies)
                {
                    if ((HierarchiesLoaded < this.RowHierarchies.Count && hierarchy.UniqueName.Equals(this.RowHierarchies[HierarchiesLoaded])) ||
                        (HierarchiesLoaded < this.ColumnHierarchies.Count && hierarchy.UniqueName.Equals(this.ColumnHierarchies[HierarchiesLoaded])) ||
                        (HierarchiesLoaded < this.FilterHierarchies.Count && hierarchy.UniqueName.Equals(this.FilterHierarchies[HierarchiesLoaded])))
                    {
                        if (this.RowHierarchies.Contains(hierarchy.UniqueName))
                        {
                            IFilterViewModel fvm = pivotGrid.DataSource.CreateFilterViewModel(hierarchy);
                            pivotGrid.DataSource.Rows.Add(fvm);
                            LoadCounter++;
                        }
                        else if (this.ColumnHierarchies.Contains(hierarchy.UniqueName))
                        {
                            IFilterViewModel fvm = pivotGrid.DataSource.CreateFilterViewModel(hierarchy);
                            pivotGrid.DataSource.Columns.Add(fvm);
                            LoadCounter++;
                        }
                        else if (this.FilterHierarchies.Contains(hierarchy.UniqueName))
                        {
                            IFilterViewModel fvm = pivotGrid.DataSource.CreateFilterViewModel(hierarchy);
                            pivotGrid.DataSource.Filters.Add(fvm);
                            LoadCounter++;
                        }
                    }
                    else
                        continue;
                }
            }

            if (LoadCounter > 0)
                HierarchiesLoaded++;
        }

        void pivotGrid_LayoutLoaded(object sender, EventArgs e)
        {
            if (!this.StateLoadComplete)
            {
                StateLoadComplete = true;

                //use deferred updates to execute all operations at once
                XamPivotGrid pivotGrid = (sender as XamPivotGrid);
                pivotGrid.DataSource.DeferredLayoutUpdate = true;

                // expanding has to be threaded
                BackgroundWorker worker = new BackgroundWorker();
                worker.DoWork += (s, evt) =>
                {
                    DataSourceBase dataSource = evt.Argument as DataSourceBase;
                    //expand members for columns, rows and filters
                    for (int i = 0; i < dataSource.Rows.Count; i++)
                    {
                        IFilterViewModel filterViewModel = dataSource.Rows[i] as IFilterViewModel;

                        if (filterViewModel != null)
                        {
                            if (filterViewModel.FilterMembers != null)
                                foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                                    ExpandMembers(filterViewModel, filterMember);
                        }
                    }
                    for (int i = 0; i < dataSource.Columns.Count; i++)
                    {
                        IFilterViewModel filterViewModel = dataSource.Columns[i] as IFilterViewModel;

                        if (filterViewModel != null)
                        {
                            if (filterViewModel.FilterMembers != null)
                                foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                                    ExpandMembers(filterViewModel, filterMember);
                        }
                    }
                    foreach (IFilterViewModel filterViewModel in dataSource.Filters)
                    {
                        if (filterViewModel.FilterMembers != null)
                            foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                                ExpandMembers(filterViewModel, filterMember);
                    }

                    //select members for columns, rows and filters
                    foreach (IFilterViewModel filterViewModel in dataSource.Filters)
                    {
                        if (filterViewModel.FilterMembers != null)
                            foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                                SelectMembers(filterMember);
                    }
                    for (int i = 0; i < dataSource.Rows.Count; i++)
                    {
                        IFilterViewModel filterViewModel = dataSource.Rows[i] as IFilterViewModel;

                        if (filterViewModel != null)
                        {
                            if (filterViewModel.FilterMembers != null)
                                foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                                    SelectMembers(filterMember);
                        }
                    }
                    for (int i = 0; i < dataSource.Columns.Count; i++)
                    {
                        IFilterViewModel filterViewModel = dataSource.Columns[i] as IFilterViewModel;

                        if (filterViewModel != null)
                        {
                            if (filterViewModel.FilterMembers != null)
                                foreach (IFilterMember filterMember in filterViewModel.FilterMembers)
                                    SelectMembers(filterMember);
                        }
                    }
                };

                worker.RunWorkerCompleted += (s, evt) =>
                {
                    //stop the deferred update and force a refresh of the UI
                    pivotGrid.DataSource.ResultChanged += DataSource_ResultChanged;
                    pivotGrid.DataSource.DeferredLayoutUpdate = false;
                };

                worker.RunWorkerAsync(pivotGrid.DataSource);

            }
        }

        void DataSource_ResultChanged(object sender, System.ComponentModel.AsyncCompletedEventArgs e)
        {
            DataSourceBase dataSourceBase = (DataSourceBase)sender;
            dataSourceBase.ResultChanged -= this.DataSource_ResultChanged;

            dataSourceBase.HierarchyExpansionMode = HierarchyExpansionMode.DrillDown;
            dataSourceBase.UpdateVisualMembers(ExecutionContext.Rows);
            dataSourceBase.UpdateVisualMembers(ExecutionContext.Columns);
            dataSourceBase.HierarchyExpansionMode = HierarchyExpansionMode.SingleLevelOnly;

            dataSourceBase.RefreshGrid();
        }

        private void SelectMembers(IFilterMember filterMember)
        {
            if (filterMember.FilterMembers != null)
            {
                foreach (IFilterMember subFilterMember in filterMember.FilterMembers)
                {
                    SelectMembers(subFilterMember);
                }
            }

            if (CustomDictContains(FilterMembersToLoad, filterMember.Member.UniqueName))
                filterMember.FilterSource.IsSelected =
                    GetCustomDictItem(FilterMembersToLoad, filterMember.Member.UniqueName).Value;
        }

        private void ExpandMembers(IFilterViewModel filterViewModel, IFilterMember filterMember)
        {
            //use this event for syncronizing the otherwise async load of members
            ManualResetEvent manualResetEvent = new ManualResetEvent(false);
            filterViewModel.LoadFilterMembersCompleted += (sender1, e1) => manualResetEvent.Set();

            if (CustomDictContains(FilterMembersToExpand, filterMember.Member.UniqueName))
                filterMember.IsExpanded =
                    GetCustomDictItem(FilterMembersToExpand, filterMember.Member.UniqueName).Value;

            if (!filterMember.IsExpanded)
                return;

            manualResetEvent.WaitOne();

            foreach (IFilterMember fm in filterMember.FilterMembers)
            {
                ExpandMembers(filterViewModel, fm);
            }

        }

        #endregion LoadMethods

        #region Helper fields and methods

        private bool StateLoadComplete;
        private bool eventAttached = false;

        private CustomDictItem GetCustomDictItem(List<CustomDictItem> ListOfItems, string UniqueName)
        {
            foreach (CustomDictItem item in ListOfItems)
            {
                if (item.UniqueName == UniqueName)
                    return item;
            }
            throw new Exception("Dict does not contain this item");
        }

        private bool CustomDictContains(List<CustomDictItem> ListOfItems, string UniqueName)
        {
            foreach (CustomDictItem item in ListOfItems)
            {
                if (item.UniqueName == UniqueName)
                    return true;
            }
            return false;
        }


        #endregion Helper fields and methods
    }

    //class to store UniqueNames and the bool values for the corresponding member
    public class CustomDictItem
    {
        public string UniqueName;
        public bool Value;

        //a parameterless constructor is needed in order to make this class serializable
        public CustomDictItem()
        {

        }

        public CustomDictItem(string UniqueName, bool Value)
        {
            this.UniqueName = UniqueName;
            this.Value = Value;
        }
    }
}
